import copy
import datetime
import logging
import os
import random
import threading
from inspect import signature
import json

import etcd
from consul_kv import Connection


class KVStore:

    # Initializes the KVStore
    def __init__(self,
                 consul_host=None,
                 consul_port=None,
                 etcd_host=None,
                 etcd_port=None,
                 kv_prefix=None,
                 reload_seconds=None,
                 reload_enabled=None,
                 initial_config={},
                 load_os_environ=True):

        self._logger = logging.getLogger('KVStore')

        self._etcd_client = None
        self._consul_kv = None
        self._kv_prefix = None

        self._reload_enabled = False

        self._config = initial_config or {}
        self._old_config = {}
        self._reload_seconds = 20
        self._on_reload_listeners = []
        self._timer_thread = None
        self._load_os_environ = load_os_environ

        # Set Defaults and Get From Environment
        consul_host = consul_host or os.getenv("CONSUL_HOST")
        consul_port = int(consul_port or os.getenv("CONSUL_PORT", 8500))
        etcd_host = etcd_host or os.getenv("ETCD_HOST")
        etcd_port = int(etcd_port or os.getenv("ETCD_PORT", 2379))
        kv_prefix = kv_prefix or os.getenv("KV_PREFIX", "/")
        reload_seconds = int(reload_seconds or os.getenv("RELOAD_CONFIG_PERIOD", 20))
        reload_enabled = bool(reload_enabled or os.getenv("RELOAD_ENABLED", False))

        self._logger.info("Initializing KVStore...")

        self._reload_seconds = reload_seconds
        self._reload_enabled = reload_enabled

        self._kv_prefix = kv_prefix
        if consul_host:
            url = "http://{0}:{1}/v1/".format(consul_host, consul_port)
            self._logger.info("Making connection to Consul at {0}:{1}, with prefix {2}"
                              .format(url, str(consul_port), kv_prefix))
            self._consul_kv = Connection(endpoint=url)
            if kv_prefix and len(kv_prefix) > 0 and kv_prefix[0] == "/":
                self._kv_prefix = kv_prefix[1:]

        elif etcd_host:
            self._logger.info(
                "Making connection to ETCD at {0}:{1}, with prefix {2}"
                    .format(etcd_host, etcd_port, kv_prefix))
            self._etcd_client = etcd.Client(host=etcd_host, port=etcd_port)

        self._logger.info("KV Prefix {0}".format(self._kv_prefix))

        if reload_enabled:
            self._reload_configuration()
        else:
            self.read_config()

    # Add an on_reload callback
    def on_reload(self, on_reload_callback):
        self._logger.debug("on_reload_callback: {0}".format(str(on_reload_callback)))
        self._on_reload_listeners.append(on_reload_callback)
        self._logger.debug("_on_reload_listeners count: {0}".format(str(len(self._on_reload_listeners))))

    # Set the on_reload callback
    def remove_on_reload(self, on_reload_callback):
        self._logger.debug("on_reload_callback: {0}".format(str(on_reload_callback)))
        self._on_reload_listeners.remove(on_reload_callback)

    # Enables Reloading
    def enable_reloading(self, reload_seconds=20):
        self._reload_seconds = reload_seconds
        self._reload_enabled = True
        self._reload_configuration()

    # Disables Reloading
    def disable_reloading(self):
        self._reload_enabled = False

    def is_reloading_enabled(self):
        return self._reload_enabled

    # Return the timer thread if present.
    def timer_thread(self):
        return self._timer_thread

    # Read Configuration from Consul or ETCD.
    def read_config(self):
        self._logger.debug("Reloading configuration...")
        old_config = self._config
        new_config = None

        # Read from consul if present
        if self._consul_kv:
            self._logger.debug("Reading configuration from consul...")
            new_config = self._consul_kv.get_dict(self._kv_prefix)
            keys = self._kv_prefix.split("/")
            for key in keys:
                self._logger.debug("Reading {0} from consul..."
                                   .format(key))
                temp_config = new_config.get(key)
                if temp_config:
                    new_config = temp_config
                self._logger.debug("{0} = {1}"
                                   .format(key, str(new_config)))

        # Read from etcd if present
        elif self._etcd_client:
            new_config = self._extract_etcd_config(self._kv_prefix)

        new_config = new_config or {}
        old_config = old_config or {}

        # Read from os.environ
        if self._load_os_environ:
            for key in os.environ.keys():
                self._logger.debug("Should I add {0},{1}={2}".format(key, key.lower(), os.environ.get(key)))
                if key not in new_config and key.lower() not in new_config:
                    self._logger.debug("Adding from environment {0},{1}={2}".format(key, key.lower(), os.environ.get(key)))
                    value = os.environ.get(key)
                    if "{" and "}" in value:
                        try:
                            value = json.loads(value)
                        except ValueError:
                            self._logger.debug("value is not JSON: {0}".format(str(value)))

                    new_config[key] = value
                    new_config[key.lower()] = value

        self._config = new_config
        self._old_config = old_config

        self._logger.debug("old_config is {0}."
                           .format(str(self._old_config)))
        self._logger.debug("new_config is {0}."
                           .format(str(self._config)))
        self._logger.debug("Configuration reloaded.")

        self._notify_on_reload_listeners(self._old_config, self._config)

    def _notify_on_reload_listeners(self, old_config, new_config):
        # Notify the on-reload listeners.
        for listener in self._on_reload_listeners:
            if listener and callable(listener):
                if len(signature(listener).parameters) == 2:
                    self._logger.debug("Notifying on_reload callback method: {0}".format(str(listener)))
                    threading.Thread(target=listener(old_config, new_config),
                                     name="KVStoreNotifyThread-{0}-{1}"
                                     .format(datetime.datetime.now().strftime("%d-%m-%Y %H:%M:%S"),
                                             self._random_string())).start()
                else:
                    self._logger.error("Expected 2 parameters on on_reload function.")
            elif listener:
                self._logger.error("Listener is not a callable function.")

    def _random_string(self):
        return str(random.randrange(0, 999999)).zfill(6)

    # Watch the configuration
    def _reload_configuration(self):
        self._logger.debug("Reloading KVStore {0}"
                           .format(str(self)))

        # Load the config and notify the listeners.
        self.read_config()

        # Starts a thread to reload the configuration periodically, this needs to happen before reloading the config,
        # in case the notify function blocks.
        if self._reload_enabled:
            self._logger.debug("Reloading configuration every {0} seconds.".format(self._reload_seconds))
            self._timer_thread = threading.Timer(self._reload_seconds, self._reload_configuration)
            self._timer_thread.daemon = True
            self._timer_thread.start()
        else:
            self._timer_thread = None


    # Gets all attributes for an etcd config given a root node.
    def _extract_etcd_config(self, etcd_path):
        self._logger.debug("Reading config data from etcd at {0}..."
                           .format(etcd_path))
        local_config = {}
        etcd_config = self._etcd_client.read(etcd_path, recursive=True, sorted=True)
        for child in etcd_config.children:
            self._logger.debug("Reading {0}..."
                               .format(str(child)))

            key_start = len(etcd_path)
            if not etcd_path.endswith("/"):
                key_start = key_start + 1
            self._logger.debug("Full key is {0}, root is {1}".format(child.key, etcd_path))
            self._set_attribute(local_config, child.key[key_start:], child.value)
        return local_config

    # Sets a deep attribute on a dict given an attribute delimited with a /
    def _set_attribute(self, a_dict, key, value):
        self._logger.debug("Setting key {0} to value {1}".format(key, value))
        last_dict = a_dict
        i = 0
        keys = key.split("/")
        for iter_key in keys:
            i = i + 1
            if i == len(keys):
                last_dict[iter_key] = value
            else:
                prev_last_dict = last_dict
                last_dict = last_dict.get(iter_key)
                if not last_dict:
                    last_dict = {}
                    prev_last_dict[iter_key] = last_dict

    # Returns true if the config attribute is different between the old and new value.
    def attribute_changed_in_dict(self, old_dict, new_dict, *keys):
        if not new_dict and not old_dict:
            self._logger.warning("KV Prefix {0}".format(self._kv_prefix))
            self._logger.warning("New and old configurations are None types.")
            return False
        if not new_dict or not old_dict:
            self._logger.warning("KV Prefix {0}".format(self._kv_prefix))
            self._logger.warning("New or old configurations are None types. old config = {0}, new config = {1}"
                                 .format(str(old_dict), str(new_dict)))
            return True

        for key in keys:
            if str(new_dict.get(key)) != str(old_dict.get(key)):
                self._logger.debug("Key {0} changed.".format(key))
                return True
        self._logger.debug("Values of {0} did NOT change.".format(str(keys)))
        return False

    # Returns true if the config attribute is different between the old and new value.
    def attribute_changed(self, *keys):
        if not self._config and not self._old_config:
            self._logger.warning("KV Prefix {0}".format(self._kv_prefix))
            self._logger.warning("New and old configurations are None types.")
            return False
        if not self._config or not self._old_config:
            self._logger.warning("KV Prefix {0}".format(self._kv_prefix))
            self._logger.warning("New or old configurations are None types. old config = {0}, new config = {1}"
                                 .format(str(self._old_config), str(self._config)))
            return True

        for key in keys:
            if str(self._config.get(key)) != str(self._old_config.get(key)):
                self._logger.debug("Key {0} changed.".format(key))
                return True
        self._logger.debug("Values of {0} did NOT change.".format(str(keys)))
        return False

    def get(self, key, default=None):
        value = None
        if key:
            keys = key.split("/")
            value = self._config
            for k in keys:
                if value and type(value) is dict:
                    value = value.get(k)
                else:
                    value = None
                    break

        return copy.deepcopy(value or default)

    def get_dict(self, default=None):
        return copy.deepcopy(self._config) or default

    def get_old_dict(self, default=None):
        return copy.deepcopy(self._old_config) or default



import logging


class LogManager:
    _DEFAULT_LOG_FORMAT = "[%(threadName)s] %(asctime)s - %(levelname)s - %(name)s - %(message)s"
    _DEFAULT_DATE_FORMAT = "%m/%d/%Y %I:%M:%S %p"

    def __init__(self,
                 kv_store):
        self._logger = logging.getLogger("LogManager")
        self._kv_store = kv_store
        logging.basicConfig(format=self._DEFAULT_LOG_FORMAT)
        self._logger.setLevel(logging.DEBUG)
        self._configure_logging(None, kv_store.get_dict(default={}))
        self._logger.setLevel(logging.INFO)
        self._kv_store.on_reload(self._configure_logging)

    def _configure_logging(self, old_config, new_config):
        self._logger.debug("BEGIN _configure_logging")

        if not old_config or not old_config == new_config:
            self._logger.debug("Configuration changed, configuring logging...")
            # Basic configuration
            basic_config = {
                "filename": new_config.get('logging_filename'),
                "filemode": new_config.get('logging_filemode'),
                "format": new_config.get('logging_format', self._DEFAULT_LOG_FORMAT),
                "datefmt": new_config.get('logging_datefmt', self._DEFAULT_DATE_FORMAT),
                "level": logging.INFO
            }
            try:
                if 'logging_level' in new_config:
                    numeric_level = getattr(logging, new_config.get('logging_level'))
                    basic_config['level'] = numeric_level
                    logging.getLogger('').setLevel(numeric_level)
                else:
                    logging.getLogger('').setLevel(logging.INFO)
            except:
                self._logger.warning("Invalid level, {1}, for root logging configuration."
                                     .format("logging_level", new_config.get('logging_level')))

            self._logger.debug("Setting base configuration to {0}"
                               .format(str(basic_config)))
            logging.basicConfig(**basic_config)

            # Specific logger configuration
            if 'logging_config' in new_config and type(new_config.get('logging_config')) is dict:
                self._logger.debug("logging_config found.")
                logging_config = new_config.get('logging_config')

                for log_name in new_config.get('logging_config'):
                    logger = logging.getLogger(log_name)
                    current_logging_config = logging_config.get(log_name)
                    self._logger.debug("Configuration for {0} is {1}.".format(log_name, str(current_logging_config)))

                    if type(current_logging_config) is str:
                        self._logger.debug("Configuring logger for {0}..."
                                           .format(log_name))
                        try:
                            numeric_level = getattr(logging, current_logging_config)
                            self._logger.debug("Configuring logging level to {0} for {1}..."
                                               .format(str(numeric_level), log_name))
                            logger.setLevel(numeric_level)
                        except:
                            self._logger.warning("Invalid level, {1}, for {0} logging configuration."
                                                 .format(log_name, current_logging_config))

            else:
                self._logger.debug("logging_config NOT found.")

        else:
            self._logger.debug("Configuration unchanged.")
        self._logger.debug("END _configure_logging")



# http://python-packaging.readthedocs.io/en/latest/minimal.html
from setuptools import setup
import os

version = "0.12.0.beta.1"
def readme():
    with open('README.rst') as f:
        return f.read()

setup(
    name='clc_msa_utils',
    version=version,
    description='CenturyLink Managed Services Anywhere Python Utilities',
    url='https://github.com/CenturyLinkCloud/mos-pyutils',
    long_description=readme(),
    author='CenturyLink Managed Services Team',
    author_email='managedos@ctl.io',
    license='MIT',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='CenturyLink ManagedServices Utilities',
    packages=['clc_msa_utils', 'clc_msa_utils.clients'],

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['python-etcd', 'consul_kv', 'pika==0.12.0'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    # extras_require={
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    # package_data={
    #     'sample': ['package_data.dat'],
    # },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    # entry_points={
    #     'console_scripts': [
    #         'sample=sample:main',
    #     ],
    # }
)
